open Tk
open Utils

let error msg =
  raise (Invalid_argument msg)
;;


(*******************************************************************

	Tk preparations

*******************************************************************)

let top = openTk () ;;

Wm.minsize_set top 1 1;; 


let cm foo = Tk.Centimeters (foo) and
    pix foo = Tk.Pixels (foo);;

let canv = Canvas.create top [Tk.Width (pix 1000); Tk.Height (pix 1000)];;

pack [canv] [] ;;

let rec ptext text x y =
  Canvas.create_text canv (pix x) (pix y) 
    [Tk.Text text; Tk.Font "-*-lucida-medium-r-*-*-*-140-*-*-*-*-*-*";
     Tk.Anchor Tk.Center]
and aligned_text text (x,y) anchor =
  Canvas.create_text canv (pix x) (pix y) 
    [Tk.Text text; Tk.Font "-*-lucida-medium-r-*-*-*-140-*-*-*-*-*-*";
     Tk.Anchor anchor]
and bbox item = Canvas.bbox canv [item]
and clear_canv () = Canvas.delete canv [Tk.Tag "all"]
and draw_line (x1, y1) (x2, y2) =
  Canvas.create_line canv [(pix x1); (pix y1);(pix x2);(pix y2)] []

and draw_box (x1, y1) (x2, y2) =
  Canvas.create_rectangle canv (pix x1) (pix y1) (pix x2) (pix y2) []

and draw_point (x1, y1) = 
  Canvas.create_oval canv (pix (x1-1)) (pix (y1-1)) 
    (pix (x1+1)) (pix (y1+1)) []
  
(* Provide for absolute positioning of a canvas item given the
   upper left corner coord. of its bounding box *)
and position item x y =
  let bbox_coords = Canvas.bbox canv [item] and 
      el1 (a,b,c,d) = a  and
      el2 (a,b,c,d) = b in
  let adj_x = x - (el1 bbox_coords) and
      adj_y = y - (el2 bbox_coords) in
  Canvas.move canv item  (pix adj_x) (pix adj_y) 

(* Draw the bounding box of an item *)
and rect item = let bb = bbox item in
match bb with
  (a,b,c,d) -> Canvas.create_rectangle canv (pix a) (pix b) (pix c)
      (pix d) []
;;


(*******************************************************************

  Graphing functions.

*******************************************************************)

(* Return the minimum and maximum values in an array.*)
let get_min_max arr = 
  let omin = ref 0.0 and omax = ref 0.0 in
  let minmax x = omin := min x !omin ;  omax := max x !omax in 
  Array.iter minmax arr ; (!omin, !omax)
;;

(* Check for NaN and Inf 
   All drawing functions w.r.t. the physical display
   *)

(* bounds are used for boxes in the world coordinate system;
   size is used for boxes in device coordinates
   *)
type  bounds = {a: float; b: float } and
      size = {x: float; y: float }
;;

(* Convert coordinates (xin, yin) in the region (xbd, ybnd) to the
   corresponding coordinates in the box with upper left corner offset
   and size siz.

   The input coordinates use a standard coordinate system;
   the device x coordinate increases to the right, 
   the device y coordinate increases downward.
   *)
let real_to_physical xbd  ybd  siz offset  xin yin =
  let x = truncate(siz.x *.(xin -. xbd.a)/.(xbd.b -. xbd.a) +. offset.x) and
      y = truncate(siz.y *.( (ybd.b -. yin)/.(ybd.b -. ybd.a))+.offset.y) in
  (x,y)
(* val real_to_physical :
   bounds -> bounds -> size -> size -> float -> float -> int * int = <fun>
   *)
;;

(* x-y plot with offset offs and display size siz *)
let draw_graph xcoords ycoords siz offs =
  let x_a, x_b = get_min_max xcoords and
      y_a, y_b = get_min_max ycoords in
  let map = real_to_physical {a=x_a; b= x_b} {a=y_a; b=y_b} siz offs in 
  let put_point x y = draw_point (map x y) and
      put_line (x1,y1) (x2,y2) = draw_line (map x1 y1) (map x2 y2) in 
  (* draw single vertical tickmark, center at (x,y) size len *)
  let xtic len x y = 
    let px,py = map x y in
    draw_line (px, py-len/2) (px, py+len/2) 
  (* draw single horiz. tickmark, center at (x,y) size len *)      
  and ytic len x y = 
    let px,py = map x y in
    draw_line (px-len/2, py) (px+len/2, py) 
  and xt_arr = linspace x_a x_b 10 
  and yt_arr = linspace y_a y_b 10 in
  let xt x = xtic 7 x y_a ; ()
  and yt y = ytic 7 x_a y ; () 
  and format flt = Printf.sprintf "%.2g" flt in
  let xax_label str x = 
    let (xp, yp) = map x y_a in
    aligned_text str (xp, yp+5) Tk.N 
  and  yax_label str y = 
    let (xp, yp) = map x_a y in
    aligned_text str (xp-5, yp) Tk.E 
  in
  Array.iter xt xt_arr ; Array.iter yt yt_arr ; (* Tick marks *)
  draw_box (map x_a y_a) (map x_b y_b) ;	(* bounding box *)
  xax_label (format x_a) x_a ; 
  xax_label (format x_b) x_b ; 
  yax_label (format y_a) y_a ;
  yax_label (format y_b) y_b ;
  iter2 put_point xcoords ycoords		(* graph *)
;;



(* Generate a grid adaptively, so that no adjacent y-values are more
   than ymin apart, unless adjacent x values are already closer than
   xmin 

   Args are:
   Function f, 
   x bounds,
   initial minimum number of points, n,
   minimum x-spacing xmin,
   minimal desired y spacing ymin
   
   Here, to keep the points organized in a tree, we have the option of
   using redundant calls of f, or to use a lookup table for already
   computed values.  The tree approach is probably more complicated
   than necessary;  first attempts took over an hour, and present
   the difficulties mentioned.

   Or, we can simply traverse a list and insert elements as needed;
   this is done below.
   *)
type point = float*float;;

let graph_data_adaptive f xbds n xmin ymin =
  let rec initial_pts f a b n = 
    let xarr = linspace a b n in
    let yarr = Array.map f xarr 
    and to_point x y = (x,y) in 
    Array.to_list (map2 to_point xarr yarr)
  and refine_list lis f xmin ymin =
    match lis with 
      [] -> []
    | a::[] -> [a]
    | a::b -> 
	let x1,y1 =  a and x2,y2 = List.hd b in 
	if (abs_float (y2-.y1) > ymin) && ((x2-.x1) > xmin) then
	  let xn = (x2+.x1)/.2.0 in let yn = f xn in 
	  refine_list (a::((xn,yn)::b)) f xmin ymin 
	else a::(refine_list b f xmin ymin)
  in
  refine_list (initial_pts f xbds.a xbds.b n) f xmin ymin
;;

let pt_list_to_flt_arrays ptlis =
  let xl, yl = List.split ptlis in
  let xarr = Array.of_list  xl 
  and yarr = Array.of_list  yl
  in 
  (xarr, yarr)
;;  




